/*
 * trails.js
 * For licence information, read licence.txt
 *
 * handling of gesture trails
 *
 */
var aioTrailDot;
var aioTrailCont = null;
var aioTrailCnt, aioTrailZoom;
var aioTrailX, aioTrailY, aioDocX, aioDocY;


function aioStartTrail(e) {
  var targetDoc = e.target.ownerDocument;
  if (targetDoc.contentType == "application/vnd.mozilla.xul+xml" || aioIsUnformattedXML(targetDoc)) return;
  if (targetDoc.defaultView.top instanceof Window) targetDoc = targetDoc.defaultView.top.document;
  var insertionNode = (targetDoc.documentElement) ? targetDoc.documentElement : targetDoc;
  var docBox = targetDoc.getBoxObjectFor(insertionNode);
  aioDocX = docBox.screenX;
  aioDocY = docBox.screenY;
  if (aioFxV1_0) {
     aioDocX -= targetDoc.defaultView.pageXOffset;
     aioDocY -= targetDoc.defaultView.pageYOffset;
  }

  aioTrailCont = targetDoc.createElementNS(xhtmlNS, "aioTrailContainer");
  insertionNode.appendChild(aioTrailCont);

  aioTrailDot = targetDoc.createElementNS(xhtmlNS, "aioTrailDot");
  aioTrailDot.style.width = aioTrailSize + "px";
  aioTrailDot.style.height = aioTrailSize + "px";
  aioTrailDot.style.background = aioTrailColor;
  aioTrailDot.style.border = "0px";
  aioTrailDot.style.position = "absolute";
  aioTrailDot.style.zIndex = 2147483647;
  aioTrailDot.style.opacity = aioTrailOpacity / 100;

  aioTrailZoom = 0;
  if (aioFxV3) {
     if (ZoomManager.useFullZoom) {
        var trailZoom = ZoomManager.zoom;
        aioTrailZoom = (trailZoom == 1) ? 0 : trailZoom;
     }
  }
  aioTrailX = e.screenX;
  aioTrailY = e.screenY;
  aioTrailCnt = 0;
}

function aioDrawTrail(e) {
  function appendDot(x, y) {
    if ((++aioTrailCnt & 1) || aioTrailSize == 1) {
      var dot = aioTrailDot.cloneNode(true);
      if (aioTrailZoom) {
         x = Math.floor(x / aioTrailZoom);
         y = Math.floor(y / aioTrailZoom);
      }
      dot.style.left = x + "px";
      dot.style.top = y + "px";
      aioTrailCont.appendChild(dot);
    }
  }
  if (!aioTrailCont) return;
  var xMove = e.screenX - aioTrailX;
  var yMove = e.screenY - aioTrailY;
  var xDecrement = xMove < 0 ? 1 : -1;
  var yDecrement = yMove < 0 ? 1 : -1;
  var i, currX = e.screenX - aioDocX, currY = e.screenY - aioDocY;
  if (Math.abs(xMove) >= Math.abs(yMove))
    for (i = xMove; i != 0; i += xDecrement)
      appendDot(currX - i, currY - Math.round(yMove * i / xMove));
  else
    for (i = yMove; i != 0; i += yDecrement)
      appendDot(currX - Math.round(xMove * i / yMove), currY - i);
  aioTrailX = e.screenX; aioTrailY = e.screenY;
}

function aioEraseTrail() {
  if (aioTrailCont && aioTrailCont.parentNode)
    aioTrailCont.parentNode.removeChild(aioTrailCont);
  aioTrailCont = null;
}
